#!/usr/bin/env ruby
#= lndir.rb -- Create a symlink tree.
#
#== SYNOPSIS
SYNOPSIS = <<-"SYNOP"

      dcreal-lndir.rb  X11 տ魯 lndir  Ruby ϼ̤ġ
      Ǥ.  X11  lndir ϰͿǥ쥯ȥʲΥĥ꡼¤
      ȿǤܥå󥯷ƵŪ˺ޤ.

      dcreal-lndir.rb Ϥεǽ˲ä, ignore ץˤä
      Υեޤϥǥ쥯ȥʤȤǽǤ. ܺ٤ϲ
       --ignore ץڤ --ignore-regexp ץ򻲾Ȥ
      .  ʤ, ˤΥե̵뤹褦ˤ
      , ܥץȤ IGNORE_DEFAULT  IGNORE_REGEXP_DEFAULT 
      ѿԽƤ.

      --delete ץꤹ뤳Ȥ, dcreal-lndir.rb ˤäƺ
      ǥ쥯ȥĥ꡼κԤȤǽǤ. ǥ쥯ȥ
      ȥܥå󥯰ʳΥեʤ褦ˤʤäƤ뤿
      , ä̾ե붲줬ޤ.

SYNOP

#== MAINTAINERS
MAINTAINERS="Yasuhiro MORIKAWA, Daisuke TSUKAHARA"

#== USAGE
USAGE = "
      dcreal-lndir.rb [options] srcdir

        srcdir ХѥξХѥǥܥå󥯤ޤ.
        ХѥξˤХѥǥܥå󥯤ޤ.

        If \"srcdir\" is relative path, relative symbolic links are created.
        Otherwise, absolute symbolic links are created."

#== OPTIONS
#
#please read output of 
#  % lndir.rb --help

#== VERSION
PACKAGE_NAME="dcreal"
URL='http://www.gfd-dennou.org/library/cc-env/dcreal'

CVSTAGNAME='$Name: dcreal-0_3_5-3 $'
CVSDATE='$Date: 2006/04/11 16:26:18 $'
CVSREVISION='$Revision: 1.3 $'

#== LICENCE
#
#This library is distributed under the terms of the Ruby license.
#You can freely distribute/modify this library.
#
#==COPYRIGHT 
#
COPYRIGHT =
  "Copyright (C) 2006 GFD Dennou Club"

#
#== FUTURE PLANS
#
#* force ץ dry-run ץϼ¼礹Τ dry-run
#  ͥ褹.
#

#== HISTORY
#
#These entries is generated by CVS automatically.
#So don't add new information manually.
#(But please adjust old log format to latest log format manually,
#if format gap between them causes).
#
#$Log: dcreal-lndir.rb,v $
#Revision 1.3  2006/04/11 16:26:18  morikawa
#* ץƬղä "=" ɲ.
#
#Revision 1.2  2006/04/11 15:47:16  morikawa
#* եοϤ, եɤ߹ư褦
#  .
#
#Revision 1.1  2006/04/11 14:50:28  morikawa
#* ƵŪ˥ܥå󥯤 lndir.rb  dcreal Υġ뷲
#  صۼ.
#
#
#2005-12-28  morikawa
#* IGNORE  IGNORE_REGEXP Υǥեͤ.
#
#2005-12-28  morikawa
#* synopsis . ignore-regexp ץ.
#
#2005-11-29  morikawa
#* Future Plans ɲ.
#
#2005-11-29  morikawa
#* ɽǤ̵λ --ignore-regexp ,
#  ̾ɽǤλ --ignore ץ
#  Ԥ褦ͤѹ.
#* --delete ץɲ
#* 񤭥ץ --force ɲ.
#* ưåѤ --dry-run ɲ.
#* ܺɽѤ --verbose ɲ.
#
#2005-11-28  morikawa
#* optparse Ѥ褦ѹ
#
#2004-04-13  daktu32
#* Υǥ쥯ȥǤʤ褦˽
#
#2004-04-13  daktu32
#* ХѥǻꤹȼԤХ
#
#2004-04-06  daktu32
#* X  lndir 򸵤˺.
#
############################################################


############################################################
## config part

# default ignored patterns
IGNORE_DEFAULT         = %w(./ ../ . ..)
IGNORE_REGEXP_DEFAULT  = ''

SAMPLE_CONF = "dcreal-lndir.conf"

## config part end
############################################################

## required libraries

require 'pathname'
require "optparse"
require 'date'

## СʥС
UPDATE=CVSDATE.sub(/\s*\$\w+: /, '').sub(/\s*\$$/, '')
FILE_VERSION=CVSREVISION.sub(/\s*\$\w+: /, '').sub(/\s*\$$/, '')
pver_tmp=CVSTAGNAME.sub(/\s*\$\w+: /, '').sub(/\s*\$$/, '').sub(/#{PACKAGE_NAME}\-/, '').gsub(/_/, '.').chomp.strip
PACKAGE_VERSION = pver_tmp == "" ? '(Latest CVS version)' : pver_tmp

## options are parsed

opt = OptionParser.new
OPTS = {}
ARGV.options{|opt|

  opt.summary_width = 23
  opt.summary_indent = ' '*6

  opt.on( '-i=VAL', '--ignore=VAL',
                             "ܥå󥯤ʤե̾
                              ꤷޤ. ʣꤹ, 줾
                              δ֤ \",\" Ƥ.
                              : src, pub_real Ȥäǥ쥯ȥ̵
                                  % lndir.rb --ignore src,pub_real ~/lib\n"
         ){|v| OPTS[:ignore] = v.gsub(/^=/, '').split(",")}

  opt.on( '-I=VAL', '--ignore-regexp=VAL',
                             "ܥå󥯤ʤ
                              ե̾ɽǻꤷޤ.
                              ʣΥեåꤹ,
                              줾δ֤ \\| Ƥ.
                              : ĥ rb, txt Υե̵
                                  % lndir.rb --ignore-regexp \\.\\(rb\\|txt\\)$ ~/lib\n"
         ){|v| OPTS[:ignore_regexp] = v.gsub(/^=/, '')}

  opt.on( '-d', '--delete',
                             "srcdir . , lndir.rb 
                              ˤäƺ줿ǥ쥯ȥĥ꡼,
                              ʤǥ쥯ȥޤϥܥå
                              ΤߤǹΤ. 
                              ̾Υե뤬¸ߤ, 顼֤.\n"
         ){|v| OPTS[:delete] = v}

  opt.on( '-f', '--force',
                             " srcdir Ƥ, 
                              ǥ쥯ȥĥ꡼, 
                              ǥ쥯ȥĥ꡼ޤ..\n"
         ){|v| OPTS[:force] = v}

#  opt.on('--relative',
#                           " ܥå󥯤Хѥǻꤹ.\n"
#         ){|v| OPTS[:relative] = v}

  opt.on('-v', '--verbose',
                             "ǥ쥯ȥȥܥå󥯤
                              ܺ٤ɽ. --delete ץƱ
                              ꤵˤ, ꥹȤɽ\n"
         ){|v| OPTS[:verbose] = v}

  opt.on('-n', '--dry-run',
                             "ޤϺǥ쥯ȥȥܥå
                              󥯤ΥꥹȤɽޤ. ưŪ
                              --verbose ץ󤬻ꤵޤ.\n"
         ){|v| OPTS[:dryrun] = v; OPTS[:verbose] = v}

  opt.on( '-c=VAL', '--config=VAL',
                             "ץ ---ignore  --ignore-regex
                              Ϳʸ, ե뤫ɤ
                              ߤޤ. եο
                              --output-sample ˤäƽϤޤ.\n"
         ){|v| OPTS[:config] = v}

  opt.on( '-o', '--output-sample',
                             "ץ --config ɤ߹ե
                              Ϥޤ. ե̾
                              #{SAMPLE_CONF} Ǥ.\n"
         ){|v| OPTS[:output_sample] = v}

  opt.on_tail('-h', '-H', '--help', 
                             "ܺ٤ʥإפϤޤ.\n"
              ){|v| OPTS[:help] = v}

  opt.parse!
}
n = nil

## help
def help(opt)
  print <<-"EOF"

  #{File.basename($0.to_s)}: #{PACKAGE_NAME} #{PACKAGE_VERSION}

    USAGE: #{USAGE}

    OPTION: \n#{opt.to_a[1..-1].join("")}
    VERSION: 
      Version: #{FILE_VERSION}, Last Update: #{UPDATE}

    #{COPYRIGHT}

  EOF
end
private :help

def help_detail(opt)
  print <<-"EOF"

  #{File.basename($0.to_s)}: #{PACKAGE_NAME} #{PACKAGE_VERSION}

    SYNOPSTIC: #{SYNOPSIS}
    USAGE: #{USAGE}

    OPTION: \n#{opt.to_a[1..-1].join("")}
    VERSION: 
      Version: #{FILE_VERSION}, Last Update: #{UPDATE}

    REFERENCE:
      ޤĤȤ椭Ҥ,ͷ, 1999, 
      ֥Ȼظץȸ Ruby , Ƕ, 135-136.

    CAUTION:
      REFERENCE ۤܤѤޤ(;^^).

    #{COPYRIGHT}

    MAINTAINERS: #{MAINTAINERS}

EOF
end


def sample_conf
  filename = SAMPLE_CONF

  print "#{File.basename($0.to_s)}: #{PACKAGE_NAME} #{PACKAGE_VERSION}\n"
  print "  Creating #{filename} ... "
  
  sfile = open(filename, "w")
  sfile.print <<-EOF
#= #{File.basename($0.to_s)}: #{PACKAGE_NAME} #{PACKAGE_VERSION} configure file
#
#  Created:: #{Time.now.strftime("%Y/%m/%d %H:%M:%S")}
#
#== Overview
#
#This file is generate by following ruby script automatically.
#
#      #{File.expand_path($0.to_s)}
#      version "#{FILE_VERSION}"
#
#Please edit this file according to your purpose.
#And Enter following command.
#
#      #{File.basename($0.to_s)} -c #{filename}
#

# Ignored file or directory names
IGNORE         = ['./', '../', '.', '..', 'CVS', 'RCS']

# Ignored file or directory names by regular expression
IGNORE_REGEXP  = '~$|\.(bak|orig)$|^SIGEN.*html?$'

EOF
  sfile.close

  print "  done.\n"

end

## methods list

#
#  srcdir ȥǥ쥯ȥƱ, ⤷
# ȥǥ쥯ȥξ̥ǥ쥯ȥξ, ֤.
#
def include_pwd?(srcdir)
  pwdpath = Pathname.new(Dir.pwd).expand_path
  srcpath = Pathname.new(srcdir).expand_path

  pwd_array = pwdpath.to_s.split("/")

  while pwd_array.size > 0
    pwd_array[0] = "/" if pwd_array.size == 1
    return true if srcpath.to_s == File.join(pwd_array)
    pwd_array.pop
  end
  false

  # ®̩˥åˡ (֤ꤹƵѲ)
#  Pathname.glob(srcpath.join("**/")){ |dir|
#    return true if dir.expand_path == pwdpath
#  }
#  false

  # ®ø˥åˡ (˿ʤäΤǵѲ)
#  return true if pwdpath == srcpath
#  return true if pwdpath.to_s =~ /^#{srcpath.to_s}/

end

#
#  dstdir ˥ܥå󥯤ȥǥ쥯ȥʳʤ
# å. ⤷⤽ʳä, 顼֤ƽλ.
#
def only_sym_or_dir?(dstdir)
  ndir = Pathname.new(dstdir.strip.chomp)

  ndir.each_entry{ |file|
    next if file.to_s == ".." || file.to_s == "."
    fullpath = ndir.join(file)
    if fullpath.directory?
      only_sym_or_dir?(fullpath.to_s)
    elsif !fullpath.symlink?
      raise ArgumentError, "\n\n  Error : \"#{fullpath.to_s}\" is not directory or symbolic link.\n\n"
    end
  }
  true
end


#
#  dstdir Υǥ쥯ȥ. , 
# ǥ쥯ȥ꤫ܥå󥯰ʳΥե뤬줿,
# 顼֤ƽλ.
#
def deldir(dstdir, origdir)
  @@depth += 1
  ndir = Pathname.new(dstdir.strip.chomp)

  # check
  unless ndir.exist?
    raise ArgumentError, "\n\n  Error : \"#{ndir.to_s}\" is not found.\n\n"
  end

  unless ndir.directory?
    raise ArgumentError, "\n\n  Error : \"#{ndir.to_s}\" is not directory.\n\n"
  end

  # ǥ쥯ȥõ, ˤΤܥå󥯤ʤк.
  # ǥ쥯ȥʤкƵŪ˥᥽å deldir ƤӽФ. ʳ
  # Υե뤬Ĥä饨顼֤ƽλ. (ǥåƤ뤬ǰΤ)
  ndir.each_entry{ |f_or_d|
    next if f_or_d.to_s == ".." || f_or_d.to_s == "."
    fullpath = ndir.join(f_or_d)
    if fullpath.directory?
      deldir(fullpath.to_s, origdir)
    elsif fullpath.symlink?
      fullpath.unlink unless OPTS[:dryrun]
      print "  deleting #{fullpath}\n" if OPTS[:verbose]
    else
      raise ArgumentError, "\n\n  Error : \"#{dir.to_s}\" is not directory or symbolic link.\n\n"
    end
  }

  # æ
  @@depth -= 1
  print "  deleting #{ndir.to_s}\n" if OPTS[:verbose]
  Dir.rmdir(ndir.to_s) unless OPTS[:dryrun]
end


#
#  srcdir Υǥ쥯ȥ깽򥫥ȥǥ쥯ȥ˺.
#
def lndir(srcdir, origdir)
  @@depth += 1
  ndir = Pathname.new(srcdir.strip.chomp)
  relpath2dst = Pathname.pwd.relative_path_from(origdir)
  if ndir.relative?
    path2src = ndir.relative_path_from(Pathname.new("../"*@@depth))
  else
    path2src = Pathname.new(ndir)
  end

  # check
  unless ndir.exist?
    raise ArgumentError, "\n\n  Error : \"#{ndir.to_s}\" is not found.\n\n"
  end

  unless ndir.directory?
    raise ArgumentError, "\n\n  Error : \"#{ndir.to_s}\" is not directory.\n\n"
  end

  if include_pwd?(srcdir)
    raise ArgumentError, "\n\n  Error : \"#{srcdir.to_s}\" include current directory.\n\n"
  end

  # create srcdir name
  dirname = ndir.expand_path.basename
  if dirname.exist?
    raise IOError, "\n\n  Error : \"#{dirname.to_s}\" is already exist.\n\n"
  end

  print "  Directory: #{relpath2dst.join(dirname)} is created\n" if OPTS[:verbose]
  dirname.mkdir
  Dir.chdir(dirname.to_s)

  # ХѥХѥǾʬ
  if ndir.absolute?
    dir = Pathname.new(ndir)
  else
    dir = Pathname.new(File.join("..", ndir.to_s))
  end

  # ǥ쥯ȥõ, ˤΤեʤ饷ܥå
  # , ǥ쥯ȥʤкƵŪ˥᥽å lndir ƤӽФ.
  dir.each_entry{ |f_or_d|
    fullpath = dir.join(f_or_d.to_s)
    next if IGNORE.include?(f_or_d.to_s)
    next if !IGNORE_REGEXP.strip.chomp.empty? && /#{IGNORE_REGEXP}/ =~ f_or_d.to_s
    next if OPTS[:ignore] && OPTS[:ignore].include?(f_or_d.to_s)
    next if OPTS[:ignore_regexp] && /#{OPTS[:ignore_regexp]}/ =~ f_or_d.to_s

    if fullpath.directory?
      fname = f_or_d.basename
      lndir(dir.join(fname.to_s).to_s, origdir)
    else 
      if f_or_d.exist?
        raise IOError, "\n\n  Error : \"#{f_or_d.to_s}\" is alread exist.\n\n"
      end
      f_or_d.make_symlink(dir.join(f_or_d)) unless OPTS[:dryrun]
      print "  SymLink  : #{path2src.join(f_or_d.basename)} => #{relpath2dst.join(dirname, f_or_d)}\n" if OPTS[:verbose]
    end
  }

  # æ
  Dir.chdir("../")
  @@depth -= 1
  Dir.rmdir(dirname.to_s) if OPTS[:dryrun]

end

## main
if OPTS[:help] then
  help_detail(opt)
  exit
end

if OPTS[:output_sample] then
  sample_conf
  exit
end

if ARGV.size == 0 then
  help(opt)
  exit
end

if OPTS[:config] then
  print "#{File.basename($0.to_s)}: #{PACKAGE_NAME} #{PACKAGE_VERSION}\n"
  print "  Loading #{OPTS[:config]} ... \n"
  $stdout.flush

  load OPTS[:config]

  print "  .. done.\n"
else
  IGNORE         = IGNORE_DEFAULT
  IGNORE_REGEXP  = IGNORE_REGEXP_DEFAULT
end


srcdir = ARGV[0]
origdir = Pathname.new(Dir.pwd)
if OPTS[:delete]
  only_sym_or_dir?(srcdir)
  @@depth = -1
  deldir(srcdir, origdir)
end
if OPTS[:force]
  only_sym_or_dir?(File.basename(srcdir))
  @@depth = -1
  deldir(File.basename(srcdir), origdir)
end
@@depth = -1
lndir(srcdir, origdir) unless OPTS[:delete]
exit 0
